from os import environ, path
from argparse import ArgumentParser
import asyncio
import logging
import toml
from build_dashboard.model import BuildbotModel, BuildbotClient
from build_dashboard.screen import draw_screen
from build_dashboard import logger


def main():
    parser = ArgumentParser(prog='build_dashboard', description='A buildbot client')
    parser.add_argument('--unix', help='Unix domain socket to connect through', type=str)
    parser.add_argument('--username', help='Username for basic auth', type=str)
    parser.add_argument('--config', help='Config file to use', type=str)
    parser.add_argument('--protocol', help='Connection protocol (Default: http)', type=str)
    parser.add_argument('--host', help='Buildbot master hostname', type=str)
    parser.add_argument('--log', help='Writes logs to file for debugging', type=str)
    parser.add_argument('--update-interval', help='Update interval', type=str)
    args = parser.parse_args()
    
    config_file = None
    if args.config:
        config_file = args.config
    elif environ.get('HOME') != None:
        config_file = environ.get('HOME') + '/.buildbotrc'

    config = {}

    if (config_file is not None and
            path.exists(config_file)):
        with open(config_file) as f:
            config.update(toml.load(f))

    for key in vars(args):
        value = getattr(args, key)
        if value != None:
            config[key] = value
    
    if args.username is not None and 'password' not in config:
        config['password'] = input("Password: ")
    
    if 'log' in config:
        handler = logging.FileHandler(config['log'])
        formatter = logging.Formatter(
            '%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)
    else:
        handler = logging.StreamHandler()
        formatter = logging.Formatter(
            '%(levelname)-8s %(message)s')
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(logging.ERROR)

    client = BuildbotClient(
            path=config.get('unix', None), 
            protocol=config.get('protocol', 'http'),
            host=config.get('host', 'localhost'),
            username=config.get('username', None),
            password=config.get('password', None)
    )
    
    model = BuildbotModel(client)
    
    loop = asyncio.get_event_loop()
    draw_screen(model, loop, config.get('update-interval', 5))



from asciimatics.scene import Scene
from asciimatics.screen import Screen
from asciimatics.exceptions import ResizeScreenError, StopApplication
from build_dashboard.views import BuildersView, BuildsView, StepsView, LogView
from build_dashboard import logger

def draw_screen(model, loop, update_secs=5):
    screen = Screen.open()
    scenes = []
    scenes.append(Scene([BuildersView(screen, model)], -1, name="BuildersView"))
    scenes.append(Scene([BuildsView(screen, model)], -1, name="BuildsView"))
    scenes.append(Scene([StepsView(screen, model)], -1, name="StepsView"))
    scenes.append(Scene([LogView(screen, model)], -1, name="LogView"))
    screen.set_scenes(scenes)

    while True:
        try:
            loop.run_until_complete(model.update())
            screen.force_update()
            screen.draw_next_frame(repeat=True)
            screen.wait_for_input(update_secs)
        except KeyboardInterrupt as e:
            screen.close()
            break
        except Exception as e:
            screen.close()
            logger.error(e)
            break



"""
Module containing the model classes for buildbot_dashboard
"""
from aiohttp import TCPConnector, UnixConnector, request, ClientSession
from aiohttp import BasicAuth
from json import loads, dumps
import asyncio
from build_dashboard import logger
from cachetools import TTLCache

class BuildbotModel(object):
    """
    The Buildbot model that is provided to the user inferface. It wraps
    a Builbot client that is used for accessing the REST API.
    
    Args:
        client (:obj:`BuildbotClient`): BuildbotClient for interactions
            with the Buildbot REST API

    Attributes:
        client (:obj:`BuildbotClient`): BuildbotClient for interactions
            with the Buildbot REST API
    """

    def __init__(self, client):
        self.client = client
        self._builders = {}
        self._selected_builder = None
        self._selected_build = None
        self._selected_log = None
        self._steps = TTLCache(10, 60)
        self._logs = TTLCache(10, 60)

    def __del__(self):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.client.close())

    def select_builder(self, builderid):
        if builderid in self._builders:
            self._selected_builder = self._builders[builderid]
        else:
            self._selected_builder = None
    
    @property
    def builder(self):
        return self._selected_builder

    def select_build(self, buildid):
        """Selects the build of selected build
        """
        self._selected_build = None
        for build in self._selected_builder['builds']:
            if build['buildid'] == buildid:
                self._selected_build = build
                break

    @property
    def build(self):
        return self._selected_build

    @property
    def steps(self):
        if self._selected_build is not None:
            buildid = self._selected_build.get('buildid')
            if buildid is not None:
                try:
                    return self._steps[buildid]
                except KeyError:
                    loop = asyncio.get_event_loop()
                    steps = loop.run_until_complete(self.client.get_steps(buildid))
                    self._steps[buildid] = steps
                    return steps
            else:
                logger.debug('Build id not found.')
        else:
            return []

    async def __mergeBuilderAndBuilds(self, builder):
        builds = await self.client.builds(builder['builderid'])
        builder['builds'] = builds 
        return builder
    
    def select_log(self, stepid):
        self._selected_log = None
        try:
            self._selected_log = self._logs[stepid]
        except KeyError:
            loop = asyncio.get_event_loop()
            self._selected_log = loop.run_until_complete(self.client.get_logs(stepid))

    @property
    def log(self):
        return self._selected_log

    def run_force_scheduler(self, builderid):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self.client.run_force_scheduler(builderid)) 

    async def update(self):
        """Performs a single update to the model
        """
        logger.debug('Updating model')
        builders = await self.client.builders()
        done, pending = await asyncio.wait([self.__mergeBuilderAndBuilds(builder) 
                    for builder in builders ])
        self._builders = { task.result()['builderid']:task.result() for task in done }
    
    def builders(self):
        """Get cached builders
        """
        return list(self._builders.values());

    def builds(self, builderid):
        """Get cached builds for builders

        Args:
            builderid (int): The id of the builder for which to retrieve builds.
        """
        return self._builders['builderid']['builds']

class BuildbotClient(object):
    """
    The Buildbot HTTP client used for accessing the REST API of Buildbot.
    
    Note:
        This client assumes the Buildbot is using API v2.

    Args:
        path (str, optional): The path to the UNIX domain socket, if using one.
        protocol (str, optional): The protocol of the REST API. Defaults to http.
        host (str, optional): The hostname of the REST API. Defaults to localhost.

    Attributes:
        session (:obj:`ClientSession`): Connection session to Buildbot REST API
        base_address (str): Base URI of the REST API.
    """
    def __init__(self, path=None, protocol='http', host='localhost', username=None, password=None):

        if path is None:
            conn = TCPConnector(limit=30)
        else:
            conn = UnixConnector(path=path)
        auth = None
        if username is not None or password is not None:
            auth = BasicAuth(login=username, password=password)
        self.session = ClientSession(connector=conn, auth=auth) 

        self.base_address = protocol + '://' + host + '/api/v2'

    async def get_steps(self, buildid):
        results = await self._get('/builds/' + str(buildid) + '/steps')
        return results['steps']
    
    async def get_logs(self, stepid):
        results = await self._get('/steps/' + str(stepid) + '/logs')
        content = ''
        if 'logs' in results:
            logids = [ log['logid'] for log in results['logs'] ]
            done, notdone = await asyncio.wait([
                self._get('/logs/' + str(logid) + '/contents') 
                for logid in logids])

            contents = [ logchunk['content'] 
                    for result in done 
                    for logchunk in result.result()['logchunks'] ]
            content = "".join(contents)
        return content

    async def get_force_schedulers(self, builderid):
       results = await self._get('/builders/' + str(builderid) + '/forceschedulers')
       return results['forceschedulers']

    async def run_force_scheduler(self, builderid):
        results = await self.get_force_schedulers(builderid)
        result = ''
        if any(results):
            name = results[0]['name']
            await self._post_jsonrpc('/forceschedulers/' + str(name), body={'method': 'force', 'params': {}})
            result = ', '.join(results[0]['builder_names'])
        return result

    async def _post_jsonrpc(self, address, body):
        message = {
            'jsonrpc': '2.0',
            'id': 1
        }
        return await self._post(address, body={**message, **body})

    async def _post(self, address, body):
        response = await self.session.post(self.base_address + address, json=body)
        text = await response.text()
        result = loads(text)
        return result

    async def _get(self, address):
        """ A template for asynchronous gets to REST API 
        
        Args:
            address (str): The relative path to the API requested

        Returns:
            A :obj:`dict` representing the response.
        """
        response = await self.session.get(self.base_address + address)
        if response.status == 401:
            raise RuntimeError("Unauthorized. Please provide valid username and password.")
        elif response.status == 403:
            raise RuntimeError("Forbidden")
        text = await response.text()
        result = loads(text)
        return result
   
    async def builders(self):
        """ Requests builders endpoint from Buildbot REST API
        
        Returns:
            A :obj:`dict` representing the response.
        """
        results = await self._get('/builders')
        return results['builders']

    async def builds(self, builderid):
        """ Requests build endpoint of a particular build from Buildbot REST API.
        
        Args:
            builderid (str): The id of the builder

        Returns:
            A :obj:`dict` representing the response.
        """
        results = await self._get('/builders/' + str(builderid) + '/builds')
        return results['builds']
    
    async def close(self):
        """ Closes the underlying :obj:`ClientSession` """
        await self.session.close()



from asciimatics.widgets import Frame, Layout, Label, Divider, MultiColumnListBox, ListBox, PopUpDialog
from asciimatics.effects import Print
from asciimatics.event import KeyboardEvent
from asciimatics.exceptions import NextScene
from asciimatics.screen import Screen
from datetime import datetime
from build_dashboard import logger

class BuildersView(Frame):
    """ Entrypoint view of the build-dashboard.

    Args:
        screen: The screen used for displaying the view.
        model: The BuildbotModel for retrieving data.
    """

    def __init__(self, screen, model):
        super(BuildersView, self).__init__(screen, screen.height, screen.width, can_scroll=False)
        self.set_theme("monochrome")
        self.model = model
        layout = Layout([100], fill_frame=True)
        self.add_layout(layout)
        layout.add_widget(Label("Buildbot"))
        layout.add_widget(Divider())
        self.builder_list = MultiColumnListBox(screen.height-4,
            columns=["20%", "30%", "15%", "35%"],
            options=[],
            on_select=self._select,
            titles=['Builder', 'Description', 'Last Build', 'Status'],
            name='builder')
        layout.add_widget(self.builder_list)
        self.fix()

    def _select(self):
        self.save()
        self.model.select_builder(self.data['builder'])
        raise NextScene(name='BuildsView')
    
    def update(self, frame):
        builders = [ BuildersView.format_builder_info(builder) 
            for builder in self.model.builders() ]
        builders.sort(key = lambda obj : obj[1])
        logger.debug("Found %s builder.", len(builders))
        self.builder_list.options = builders
        self.model.select_builder(builders[0][1])
        Frame.update(self, frame) 

    def process_event(self, event):
        if (event is not None and
                isinstance(event, KeyboardEvent)):
            if event.key_code == ord('f') or event.key_code == ord('F'):
                self.save()
                builders = self.model.run_force_scheduler(self.data['builder'])
                self._scene.add_effect(
                        PopUpDialog(self._screen, "Forced build of: %s" % (builders), ["OK"]))
        return super(BuildersView, self).process_event(event)
        
    @staticmethod
    def format_builder_info(builder):
        """ Formats the merged builder and builds message into the columns
        for a :obj:`MultiColumnListBox` in :obj:`BuildersView`.

        Args:
            builder (:obj:`dict`): A builder :obj:`dict` with the merged
                builds :obj:`dict`.

        Returns:
            A :obj:`tuple` with four columns and id.
        """
        name = builder['name'] or 'None'
        description = builder['description'] or 'None'
        last_build_time = ''
        state_string = ''
        if any(builder['builds']):
            last_build = builder['builds'][-1]
            if last_build['complete']:
                last_build_time = datetime.utcfromtimestamp(
                            last_build['complete_at']).strftime(
                                    '%Y-%m-%d %H:%M:%S')
            state_string = last_build['state_string']
        builderid = builder['builderid']
        formatted = ([name,
            description,
            last_build_time,
            state_string],
            builderid)
        return formatted

class BuildsView(Frame):
    """Frame to display the list of builds for a builder

    Args:
        screen (:obj:`Screen`): The screen object
        model: The Buildbot model
    """
    def __init__(self, screen, model):
        super(BuildsView, self).__init__(screen, screen.height, screen.width, can_scroll=False)
        self.set_theme("monochrome")
        self.model = model
        layout = Layout([100], fill_frame=True)
        self.add_layout(layout)
        self.builds = []
        layout.add_widget(Label("Builds"))
        layout.add_widget(Divider())
        layout.add_widget(MultiColumnListBox(screen.height - 4,
            columns=["20%", "15%", "15%", "50%"],
            options=self.builds,
            on_select=self._select,
            titles=['Number', 'Started At', 'Completed At', 'Status'],
            name='build'))
        self.fix()

    def update(self, frame):
        if self.model.builder:
            builds = [ BuildsView.format_build_info(builder) 
                for builder in self.model.builder.get('builds',[]) ]
            logger.debug("Found %s builds.", len(builds))
            self.builds.clear()
            self.builds.extend(builds)
        Frame.update(self, frame)
    
    def _select(self):
        self.save()
        self.model.select_build(self.data['build'])
        raise NextScene(name="StepsView")
    
    def process_event(self, event):
        if (event is not None and isinstance(event, KeyboardEvent)):
            if event.key_code == -1:
                raise NextScene(name="BuildersView")
        return super(BuildsView, self).process_event(event)

    @staticmethod
    def format_build_info(build):
        number = build['number']
        if build['complete']:
            complete_time = datetime.utcfromtimestamp(
                    build['complete_at']).strftime('%Y-%m-%d %H:%M:%S')
        else:
            complete_time = ''
        start_time = datetime.utcfromtimestamp(
                    build['started_at']).strftime('%Y-%m-%d %H:%M:%S')
        state_string = build['state_string']
        buildid = build['buildid']
        formatted = ([str(number), start_time, complete_time, state_string], buildid)
        return formatted

class StepsView(Frame):
    """Frame to display the list of steps for a build

    Args:
        screen (:obj:`Screen`): The screen object
        model: The Buildbot model
    """
    def __init__(self, screen, model):
        super(StepsView, self).__init__(screen, screen.height, screen.width, can_scroll=False)
        self.set_theme("monochrome")
        self.model = model
        self.steps = []
        layout = Layout([100], fill_frame=True)
        self.add_layout(layout)
        layout.add_widget(Label("Steps"))
        layout.add_widget(Divider())
        layout.add_widget(MultiColumnListBox(screen.height - 4,
                columns=["20%", "15%", "15%", "50%"],
                options=self.steps,
                on_select=self._select,
                titles=['Number', 'Started At', 'Completed At', 'Status'],
                name='step'))

        self.fix()

    def _select(self):
        self.save()
        self.model.select_log(self.data['step'])
        raise NextScene(name='LogView')

    def update(self, frame):
        if self.model.steps:
            steps = [ StepsView.format_step_info(step) 
                for step in self.model.steps ]
            logger.debug("Found %s steps.", len(steps))
            self.steps.clear()
            self.steps.extend(steps)
        Frame.update(self, frame)
    
    def process_event(self, event):
        if (event is not None and isinstance(event, KeyboardEvent)):
            if event.key_code == -1:
                raise NextScene(name="StepsView")
        return Frame.process_event(self, event)

    @staticmethod
    def format_step_info(build):
        number = build['number']
        if build['complete']:
            complete_time = datetime.utcfromtimestamp(
                    build['complete_at']).strftime('%Y-%m-%d %H:%M:%S')
        else:
            complete_time = ''
        start_time = datetime.utcfromtimestamp(
                    build['started_at']).strftime('%Y-%m-%d %H:%M:%S')
        state_string = build['state_string']
        stepid = build['stepid']
        formatted = ([str(number), start_time, complete_time, state_string], stepid)
        return formatted


class LogView(Frame):
    """Frame to display the logs

    Args:
        screen (:obj:`Screen`): The screen object
        model: The Buildbot model
    """
    def __init__(self, screen, model):
        super(LogView, self).__init__(screen, screen.height, screen.width, can_scroll=False)
        self.set_theme("monochrome")
        self.model = model
        self.log_lines = []
        layout = Layout([100], fill_frame=True)
        self.add_layout(layout)
        layout.add_widget(Label("Logs"))
        layout.add_widget(Divider())
        layout.add_widget(ListBox(
            screen.height - 4,
            options=self.log_lines,
            add_scroll_bar=True))
        self.fix()

    def update(self, frame):
        if self.model.log:
            self.log_lines.clear()
            self.log_lines.extend(LogView.format_logs(self.model.log))
        Frame.update(self, frame)
    
    def process_event(self, event):
        if (event is not None and isinstance(event, KeyboardEvent)):
            if event.key_code == -1:
                raise NextScene(name="StepsView")
        return Frame.process_event(self, event)

    @staticmethod
    def format_logs(content):
        lines = content.splitlines()
        counter = 0
        lines = [ (line[1:], index) for index, line in enumerate(lines) ]
        return lines




from setuptools import setup
import os

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as long_d_f:
    long_description = long_d_f.read()

setup(
  name = 'build-dashboard',
  packages = ['build_dashboard'],
  version = '0.2.6',
  description = 'Buildbot CLI Dashboard',
  long_description = long_description,
  long_description_content_type = 'text/x-rst',
  author = 'Jeffrey Hill',
  author_email = 'jeff@reverentengineer.com',
  url = 'https://github.com/ReverentEngineer/build-dashboard',
  keywords = ['buildbot', 'continuous integration', 'ci', 'cli' ],
  classifiers = [
      'Development Status :: 3 - Alpha',
      'Intended Audience :: Developers',
      'Topic :: Software Development :: Build Tools',
      'License :: OSI Approved :: Apache Software License',
      'Programming Language :: Python :: 3',
      ],
  entry_points={
      'console_scripts': [
            'build-dashboard = build_dashboard.cli:main'
        ]
    },
  install_requires = ['aiohttp', 'toml', 'asciimatics','cachetools'],
  setup_requires = ['pytest-runner'],
  tests_require = ['pytest']
)
